@@ -1,7 +1,7 @@
-__version__ = "0.2.1"
+__version__ = "0.3.0"
 
 from .businessrule import *
 from .rules import *
-from .nodes import *
+from .splits import *
 from .estimators import *
 from .dashboard import *
